"""
Persistence errors.

Errors define a `status_code` for each translation to HTTP (because REST)
but are not coupled with any HTTP library.

"""


class ModelIntegrityError(Exception):
    """
    An attempt to create or update a model violated a schema constraint.

    Usually the result of a programming error.

    """
    @property
    def status_code(self):
        # internal server error
        return 500


class ModelNotFoundError(Exception):
    """
    The queried or updated model did not exist.

    """
    @property
    def status_code(self):
        # not found
        return 404



"""
Create databases.

"""
from argparse import ArgumentParser

from microcosm_dynamodb.operations import create_all, drop_all


def parse_args(graph):
    parser = ArgumentParser()
    parser.add_argument("--drop", "-D", action="store_true")
    return parser.parse_args()


def main(graph):
    """
    Create and drop databases.

    """
    args = parse_args(graph)

    if args.drop:
        drop_all(graph)
    create_all(graph)



"""
Factory that configures flywheel DynamoDB ORM-like framework.

"""
from flywheel import Engine

from microcosm.api import defaults


@defaults(
    region="us-west-2",
)
def configure_flywheel_engine(graph):
    """
    Create the flywheel engine.

    """
    namespace = ()
    if graph.metadata.testing:
        namespace = 'test'

    engine = Engine(namespace=namespace)
    engine.connect_to_region(graph.config.dynamodb.region)

    return engine



"""
Support for building models.

Every model must inherit from `Model` and should inherit from the `EntityMixin`.

"""
from flywheel import Model  # noqa


class IdentityMixin(object):
    """
    Define model identity in terms of members.

    This form of equality isn't always appropriate, but it's a good place to start,
    especially for writing test assertions.

    """
    def __eq__(self, other):
        return type(other) is type(self) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__dict__)


class SmartMixin(object):
    """
    Define a model with short cuts for CRUD operations against its `Store`.

    These short cuts still delegate responsibility for persistence to the store (which must be
    instantiated first).

    """
    def create(self):
        return self.__class__.store.create(self)

    def delete(self):
        return self.__class__.store.delete(self.id)

    def update(self):
        return self.__class__.store.update(self.id, self)

    def replace(self):
        return self.__class__.store.replace(self.id, self)

    @classmethod
    def search(cls, *criterion, **kwargs):
        return cls.store.search(*criterion, **kwargs)

    @classmethod
    def count(cls, *criterion):
        return cls.store.count(*criterion)

    @classmethod
    def retrieve(cls, identifier):
        return cls.store.retrieve(identifier)


class EntityMixin(IdentityMixin, SmartMixin):
    """
    Convention for persistent entities combining other mixins.

    """
    pass



"""
Abstraction layer for persistence operations.

"""
from operator import add
import logging

from microcosm_dynamodb.errors import (
    ModelNotFoundError,
)
from microcosm_dynamodb.identifiers import new_object_id


TEST_TABLE_PREFIX = "test_"


class Store(object):

    def __init__(self, graph, model_class):
        self.graph = graph
        self.model_class = model_class
        # Give the model class a backref to allow model-oriented CRUD
        # short cuts while still having an abstraction layer we can replace.
        self.model_class.store = self
        # Each model must be registered with the engine exactly once.
        self._register()

    @property
    def engine(self):
        return self.graph.dynamodb

    def new_object_id(self):
        """
        Injectable id generation to facilitate mocking.

        """
        return new_object_id()

    def create(self, instance):
        """
        Create a new model instance.

        """
        if instance.id is None:
            instance.id = self.new_object_id()
        self.engine.save(instance)
        return instance

    def retrieve(self, identifier, *criterion):
        """
        Retrieve a model by primary key and zero or more other criteria.

        :raises `NotFound` if there is no existing model

        """
        return self._retrieve(
            self.model_class.id == identifier,
            *criterion
        )

    def update(self, identifier, new_instance):
        """
        Update an existing model with a new one.

        :raises `ModelNotFoundError` if there is no existing model

        """
        instance = self.engine.get(self.model_class, id=identifier)
        if not instance:
            raise ModelNotFoundError()

        new_instance.id = identifier
        new_instance.sync()

        return new_instance

    def replace(self, identifier, new_instance):
        """
        Create or update a model.

        """
        try:
            return self.update(identifier, new_instance)
        except ModelNotFoundError:
            return self.create(new_instance)

    def delete(self, identifier):
        """
        Delete a model by primary key.

        :raises `ModelNotFoundError` if the row cannot be deleted.

        """
        return self._delete(self.model_class.id == identifier)

    def count(self, *criterion, **kwargs):
        """
        Count the number of models matching some criterion.

        """
        if not (criterion or kwargs):
            logging.warning(
                "count() - DynamoDB Table scans are extremely slow, avoid counting without filters when possible."
            )
            return reduce(add, (1 for item in self.engine.scan(self.model_class).gen()), 0)
        else:
            query = self._query(*criterion)
            query = self._filter(query, **kwargs)
            return query.count()

    def search(self, *criterion, **kwargs):
        """
        Return the list of models matching some criterion.

        :param limit: pagination limit, if any
        """
        if criterion:
            query = self._query(*criterion)
        else:
            query = self.engine.scan(self.model_class)

        query = self._filter(query, **kwargs)
        query = self._order_by(query, **kwargs)
        return query.all()

    def _order_by(self, query, **kwargs):
        """
        Add an order by clause to a (search) query. For DynamoDB, using flywheel,
        we can support ordering using a an index, e.g updating query with:

        >>> query = query.index('timestamp-index')

        By default, is a noop.

        """
        return query

    def _filter(self, query, **kwargs):
        """
        Filter a query with user-supplied arguments.

        :param limit: pagination limit, if any

        """
        limit = kwargs.get("limit")
        if limit is not None:
            query = query.limit(limit)
        return query

    def _retrieve(self, *criterion):
        """
        Retrieve a model by some criteria.

        :raises `ModelNotFoundError` if the row cannot be deleted.

        """
        try:
            return self._query(*criterion).one()
        except Exception as error:
            raise ModelNotFoundError(error)

    def _delete(self, *criterion):
        """
        Delete a model by some criterion.

        Avoids race-condition check-then-delete logic by checking the count of affected rows.

        :raises `ResourceNotFound` if the row cannot be deleted.

        """
        count = self._query(*criterion).delete()
        if count == 0:
            raise ModelNotFoundError
        return True

    def _query(self, *criterion):
        """
        Construct a query for the model.

        """
        return self.engine.query(
            self.model_class
        ).filter(
            *criterion
        )

    def _register(self):
        self.graph.dynamodb.register(self.model_class)



"""
Example models and store usage.

"""
from flywheel import Field

from microcosm.api import binding
from microcosm_dynamodb.models import EntityMixin, Model
from microcosm_dynamodb.store import Store


class Company(EntityMixin, Model):
    """
    A company has a name.

    """
    id = Field(hash_key=True)
    name = Field()


class CompanyStore(Store):
    pass


class Employee(EntityMixin, Model):
    """
    A employee has a name and associated company.

    """
    company_id = Field(hash_key=True)
    id = Field(range_key=True)
    name = Field()


class EmployeeStore(Store):
    pass


@binding("company_store")
def configure_company_store(graph):
    return CompanyStore(graph, Company)


@binding("employee_store")
def configure_employee_store(graph):
    return EmployeeStore(graph, Company)



"""
Common database operations.

"""


def create_all(graph):
    """
    Create all database tables.

    """
    graph.dynamodb.create_schema()


def drop_all(graph):
    """
    Drop all database tables.

    """
    graph.dynamodb.delete_schema()


def recreate_all(graph):
    """
    Drop and add back all database tables.

    """
    drop_all(graph)
    create_all(graph)
